@@ -11,16 +11,23 @@ from neuralnets.util.metrics import jaccard, accuracy_metrics, hausdorff_distanc
 from neuralnets.util.tools import gaussian_window, tensor_to_device, module_to_device
 
 
-def sliding_window_multichannel(image, step_size, window_size, track_progress=False):
+def sliding_window_multichannel(image, step_size, window_size, in_channels=1, track_progress=False):
     """
     Iterator that acts as a sliding window over a multichannel 3D image
 
     :param image: multichannel image (4D array)
     :param step_size: step size of the sliding window (3-tuple)
     :param window_size: size of the window (3-tuple)
+    :param in_channels: amount of subsequent slices that serve as input for the network (should be odd)
     :param track_progress: optionally, for tracking progress with progress bar
     """
 
+    # adjust z-channels if necessary
+    window_size = np.asarray(window_size)
+    is2d = window_size[0] == 1
+    if is2d:  # 2D
+        window_size[0] = in_channels
+
     # define range
     zrange = [0]
     while zrange[-1] < image.shape[1] - window_size[0]:
@@ -43,9 +50,9 @@ def sliding_window_multichannel(image, step_size, window_size, track_progress=Fa
             for x in xrange:
 
                 # yield the current window
-                if window_size[0] == 1:  # 2D
-                    yield (z, y, x, image[:, z, y:y + window_size[1], x:x + window_size[2]])
-                else:  # 3D
+                if is2d:
+                    yield (z, y, x, image[0, z:z + window_size[0], y:y + window_size[1], x:x + window_size[2]])
+                else:
                     yield (z, y, x, image[:, z:z + window_size[0], y:y + window_size[1], x:x + window_size[2]])
                 if track_progress:
                     bar.next()
@@ -70,33 +77,52 @@ def _init_gaussian_window(input_shape, is2d):
     return g_window
 
 
-def _init_sliding_window(data, step_size, input_shape, is2d, track_progress):
+def _init_sliding_window(data, step_size, input_shape, in_channels, is2d, track_progress):
     if is2d:
         sw = sliding_window_multichannel(data, step_size=step_size, window_size=(1, input_shape[0], input_shape[1]),
-                                         track_progress=track_progress)
+                                         in_channels=in_channels, track_progress=track_progress)
     else:
         sw = sliding_window_multichannel(data, step_size=step_size, window_size=input_shape,
                                          track_progress=track_progress)
     return sw
 
 
-def _init_batch(batch_size, channels, input_shape, is2d):
-    if is2d:
-        batch = np.zeros((batch_size, channels, input_shape[0], input_shape[1]))
+def _orient(data, orientation=0):
+    """
+    This function essentially places the desired orientation axis to that of the original Z-axis
+    For example:
+          (C, Z, Y, X) -> (C, Y, Z, X) for orientation=1
+          (C, Z, Y, X) -> (C, X, Y, Z) for orientation=2
+    Note that applying this function twice corresponds to the identity transform
+
+    :param data: assumed to be of shape (C, Z, Y, X)
+    :param orientation: 0, 1 or 2 (respectively for Z, Y or X axis)
+    :return: reoriented dataset
+    """
+    if orientation == 1:
+        return np.transpose(data, axes=(0, 2, 1, 3))
+    elif orientation == 2:
+        return np.transpose(data, axes=(0, 3, 2, 1))
     else:
-        batch = np.zeros((batch_size, channels, input_shape[0], input_shape[1], input_shape[2]))
-    return batch
+        return data
 
 
-def _pad(data, input_shape):
+def _pad(data, input_shape, in_channels):
     # pad data if input shape is larger than data
     in_shape = input_shape if len(input_shape) == 3 else (1, input_shape[0], input_shape[1])
-    pad_width = [(0, 0), None, None, None]
+    pad_width = [[0, 0], None, None, None]
     for d in range(3):
         padding = np.maximum(0, in_shape[d] - data.shape[d + 1])
         before = padding // 2
         after = padding - before
-        pad_width[d + 1] = (before, after)
+        pad_width[d + 1] = [before, after]
+
+    # pad z-slices if necessary (required if the network uses more than 1 input channel)
+    if in_channels > 1:
+        c = (in_channels // 2)
+        pad_width[1][0] = pad_width[1][0] + c
+        pad_width[1][1] = pad_width[1][1] + c
+
     return np.pad(data, pad_width=pad_width, mode='symmetric'), pad_width
 
 
@@ -117,11 +143,14 @@ def _forward_prop(net, x):
     return F.softmax(outputs, dim=1)
 
 
-def _cumulate_segmentation(seg_cum, counts_cum, outputs, g_window, positions, batch_size, input_shape, is2d):
+def _cumulate_segmentation(seg_cum, counts_cum, outputs, g_window, positions, batch_size, input_shape, in_channels,
+                           is2d):
+    c = in_channels // 2
     for b in range(batch_size):
         (z_b, y_b, x_b) = positions[b, :]
         # take into account the gaussian filtering
         if is2d:
+            z_b += c  # correct channel shift
             seg_cum[:, z_b, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += \
                 np.multiply(g_window, outputs.data.cpu().numpy()[b, ...])
             counts_cum[z_b:z_b + 1, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += g_window
@@ -131,28 +160,32 @@ def _cumulate_segmentation(seg_cum, counts_cum, outputs, g_window, positions, ba
             counts_cum[z_b:z_b + input_shape[0], y_b:y_b + input_shape[1], x_b:x_b + input_shape[2]] += g_window
 
 
-def _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, is2d):
+def _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, in_channels,
+                   is2d):
     # convert to tensors and switch to correct device
     inputs = tensor_to_device(torch.FloatTensor(batch), device)
     # forward prop
     outputs = _forward_prop(net, inputs)
     # cumulate segmentation volume
-    _cumulate_segmentation(seg_cum, counts_cum, outputs, g_window, positions, batch_size, input_shape, is2d)
+    _cumulate_segmentation(seg_cum, counts_cum, outputs, g_window, positions, batch_size, input_shape, in_channels,
+                           is2d)
 
 
-def segment_multichannel(data, net, input_shape, batch_size=1, step_size=None, train=False, track_progress=False,
-                         device=0):
+def segment_multichannel(data, net, input_shape, in_channels=1, batch_size=1, step_size=None, train=False,
+                         track_progress=False, device=0, orientation=0):
     """
     Segment a multichannel 3D image using a specific network
 
     :param data: 4D array (C, Z, Y, X) representing the multichannel 3D image
     :param net: image-to-image segmentation network
     :param input_shape: size of the inputs (either 2 or 3-tuple)
+    :param in_channels: amount of subsequent slices that serve as input for the network (should be odd)
     :param batch_size: batch size for processing
     :param step_size: step size of the sliding window
     :param train: evaluate the network in training mode
     :param track_progress: optionally, for tracking progress with progress bar
     :param device: GPU device where the computations should occur
+    :param orientation: orientation to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
     :return: the segmented image
     """
 
@@ -165,15 +198,20 @@ def segment_multichannel(data, net, input_shape, batch_size=1, step_size=None, t
     else:
         net.eval()
 
-    # get the amount of channels
-    channels = data.shape[0]
+    # orient data if necessary
+    data = _orient(data, orientation)
 
     # pad data if necessary
-    data, pad_width = _pad(data, input_shape)
+    data, pad_width = _pad(data, input_shape, in_channels)
 
     # 2D or 3D
     is2d = len(input_shape) == 2
 
+    # get the amount of channels
+    channels = data.shape[0]
+    if is2d:
+        channels = in_channels
+
     # initialize the step size
     step_size = _init_step_size(step_size, input_shape, is2d)
 
@@ -185,11 +223,11 @@ def segment_multichannel(data, net, input_shape, batch_size=1, step_size=None, t
     counts_cum = np.zeros(data.shape[1:])
 
     # define sliding window
-    sw = _init_sliding_window(data, step_size, input_shape, is2d, track_progress)
+    sw = _init_sliding_window(data, step_size, input_shape, in_channels, is2d, track_progress)
 
     # start prediction
     batch_counter = 0
-    batch = _init_batch(batch_size, channels, input_shape, is2d)
+    batch = np.zeros((batch_size, channels, *input_shape))
     positions = np.zeros((batch_size, 3), dtype=int)
     for (z, y, x, inputs) in sw:
 
@@ -203,43 +241,53 @@ def segment_multichannel(data, net, input_shape, batch_size=1, step_size=None, t
         # perform segmentation when a full batch is filled
         if batch_counter == batch_size:
             # process a single batch
-            _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, is2d)
+            _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape,
+                           in_channels, is2d)
 
             # reset batch counter
             batch_counter = 0
 
     # don't forget to process the last batch
-    _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, is2d)
+    _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, in_channels,
+                   is2d)
 
     # crop out the symmetric extension and compute segmentation
     data, seg_cum, counts_cum = _crop(data, seg_cum, counts_cum, pad_width)
     for c in range(net.out_channels):
         seg_cum[c, ...] = np.divide(seg_cum[c, ...], counts_cum)
 
+    # reorient data to its original orientation
+    data = _orient(data, orientation)
+    seg_cum = _orient(seg_cum, orientation)
+
     return seg_cum
 
 
-def segment(data, net, input_shape, batch_size=1, step_size=None, train=False, track_progress=False, device=0):
+def segment(data, net, input_shape, in_channels=1, batch_size=1, step_size=None, train=False, track_progress=False,
+            device=0, orientation=0):
     """
     Segment a 3D image using a specific network
 
     :param data: 3D array (Z, Y, X) representing the 3D image
     :param net: image-to-image segmentation network
-    :param input_shape: size of the inputs (either 2 or 3-tuple)
+    :param input_shape: size of
+    :param in_channels: Amount of subsequent slices that serve as input for the network (should be odd)
     :param batch_size: batch size for processing
     :param step_size: step size of the sliding window
     :param train: evaluate the network in training mode
     :param track_progress: optionally, for tracking progress with progress bar
     :param device: GPU device where the computations should occur
+    :param orientation: orientation to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
     :return: the segmented image
     """
 
-    return segment_multichannel(data[np.newaxis, ...], net, input_shape, batch_size=batch_size, step_size=step_size,
-                                train=train, track_progress=track_progress, device=device)
+    return segment_multichannel(data[np.newaxis, ...], net, input_shape, in_channels=in_channels, batch_size=batch_size,
+                                step_size=step_size, train=train, track_progress=track_progress, device=device,
+                                orientation=orientation)
 
 
-def validate(net, data, labels, input_size, classes_of_interest=(0, 1), batch_size=1, write_dir=None, val_file=None,
-             writer=None, epoch=0, track_progress=False, device=0):
+def validate(net, data, labels, input_size, in_channels=1, classes_of_interest=(0, 1), batch_size=1, write_dir=None,
+             val_file=None, writer=None, epoch=0, track_progress=False, device=0, orientations=(0, )):
     """
     Validate a network on a dataset and its labels
 
@@ -247,6 +295,7 @@ def validate(net, data, labels, input_size, classes_of_interest=(0, 1), batch_si
     :param data: 3D array (Z, Y, X) representing the 3D image
     :param labels: 3D array (Z, Y, X) representing the 3D labels
     :param input_size: size of the inputs (either 2 or 3-tuple) for processing
+    :param in_channels: Amount of subsequent slices that serve as input for the network (should be odd)
     :param classes_of_interest: index of the label of interest
     :param batch_size: batch size for processing
     :param write_dir: optionally, specify a directory to write the output
@@ -255,6 +304,7 @@ def validate(net, data, labels, input_size, classes_of_interest=(0, 1), batch_si
     :param epoch: optionally, current epoch for logging to tensorboard
     :param track_progress: optionally, for tracking progress with progress bar
     :param device: GPU device where the computations should occur
+    :param orientations: list of orientations to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
     :return: validation results, i.e. accuracy, precision, recall, f-score, jaccard and dice score
     """
 
@@ -263,8 +313,12 @@ def validate(net, data, labels, input_size, classes_of_interest=(0, 1), batch_si
     if write_dir is not None and not os.path.exists(write_dir):
         os.mkdir(write_dir)
 
-    # compute segmentation
-    segmentation = segment(data, net, input_size, batch_size=batch_size, track_progress=track_progress, device=device)
+    # compute segmentation for each orientation and average results
+    segmentation = np.zeros((net.out_channels, *data.shape))
+    for orientation in orientations:
+        segmentation += segment(data, net, input_size, in_channels=in_channels, batch_size=batch_size,
+                                track_progress=track_progress, device=device, orientation=orientation)
+    segmentation = segmentation / len(orientations)
 
     # compute metrics
     w = labels != 255
@@ -280,7 +334,7 @@ def validate(net, data, labels, input_size, classes_of_interest=(0, 1), batch_si
             h = -1
 
         # report results
-        print('[%s] Validation performance for class %d: ' % (datetime.datetime.now(), i))
+        print('[%s] Validation performance for class %d: ' % (datetime.datetime.now(), classes_of_interest[i]))
         print('[%s]     - Accuracy: %f' % (datetime.datetime.now(), ams[i - 1][0]))
         print('[%s]     - Balanced accuracy: %f' % (datetime.datetime.now(), ams[i - 1][1]))
         print('[%s]     - Precision: %f' % (datetime.datetime.now(), ams[i - 1][2]))
@@ -291,7 +345,7 @@ def validate(net, data, labels, input_size, classes_of_interest=(0, 1), batch_si
 
         # write stuff if necessary
         if write_dir is not None:
-            print('[%s] Writing the output for class %d: ' % (datetime.datetime.now(), i))
+            print('[%s] Writing the output for class %d: ' % (datetime.datetime.now(), classes_of_interest[i]))
             wdir = os.path.join(write_dir, 'class_' + str(i))
             if not os.path.exists(wdir):
                 os.mkdir(wdir)
